@@ -1,5 +1,5 @@
 # Copyright (C) 2013 Ishraq Ibne Ashraf <ishraq@tinkerforge.com>
-# Copyright (C) 2014 Matthias Bolte <matthias@tinkerforge.com>
+# Copyright (C) 2014, 2020 Matthias Bolte <matthias@tinkerforge.com>
 #
 # Redistribution and use in source and binary forms of this file,
 # with or without modification, are permitted. See the Creative
@@ -248,6 +248,7 @@ sub new
 	                                 auto_reconnect_allowed => 0,
 	                                 auto_reconnect_pending => 0,
 	                                 devices => shared_clone({}),
+	                                 replace_lock_ref => undef, # used to synchronize replacements in the devices hash
 	                                 registered_callbacks => shared_clone({}),
 	                                 socket_fileno => undef, # protected by SOCKET_LOCK
 	                                 socket_id => 0,
@@ -269,12 +270,14 @@ sub new
 
 	bless($self, $class);
 
+	my $replace_lock :shared;
 	my $socket_lock :shared;
 	my $local_socket_lock :shared;
 	my $send_lock :shared;
 	my $sequence_number_lock :shared;
 	my $authentication_lock :shared;
 
+	$self->{replace_lock_ref} = \$replace_lock;
 	$self->{socket_lock_ref} = \$socket_lock;
 	$self->{local_socket_lock_ref} = \$local_socket_lock;
 	$self->{send_lock_ref} = \$send_lock;
@@ -699,20 +702,22 @@ sub _brickd_create
 	$self->{brickd} = Tinkerforge::Device->_new('2', $self, [2, 0, 0]);
 	$self->{brickd}->{response_expected}->{&_BRICK_DAEMON_FUNCTION_GET_AUTHENTICATION_NONCE} = Tinkerforge::Device->_RESPONSE_EXPECTED_ALWAYS_TRUE;
 	$self->{brickd}->{response_expected}->{&_BRICK_DAEMON_FUNCTION_AUTHENTICATE} = Tinkerforge::Device->_RESPONSE_EXPECTED_TRUE;
+
+	$self->_add_device($self->{brickd});
 }
 
 sub _brickd_get_authentication_nonce
 {
 	my ($self) = @_;
 
-	return $self->{brickd}->_send_request(&_BRICK_DAEMON_FUNCTION_GET_AUTHENTICATION_NONCE, [], '', 'C4');
+	return $self->{brickd}->_send_request(&_BRICK_DAEMON_FUNCTION_GET_AUTHENTICATION_NONCE, [], '', 12, 'C4');
 }
 
 sub _brickd_authenticate
 {
 	my ($self, $clientNonce, $digest) = @_;
 
-	$self->{brickd}->_send_request(&_BRICK_DAEMON_FUNCTION_AUTHENTICATE, [$clientNonce, $digest], 'C4 C20', '');
+	$self->{brickd}->_send_request(&_BRICK_DAEMON_FUNCTION_AUTHENTICATE, [$clientNonce, $digest], 'C4 C20', 0, '');
 }
 
 =item authenticate()
@@ -991,6 +996,20 @@ sub _handle_disconnect_by_peer
 	                                  $disconnect_reason, $socket_id]);
 }
 
+sub _add_device
+{
+	my ($self, $device) = @_;
+
+	lock(${$self->{replace_lock_ref}});
+
+	if (defined($self->{devices}->{$device->{uid}}))
+	{
+		$self->{devices}->{$device->{uid}}->{replaced} = 1;
+	}
+
+	$self->{devices}->{$device->{uid}} = $device;
+}
+
 sub _ipcon_send
 {
 	my ($self, $packet) = @_;
@@ -1524,7 +1543,7 @@ sub _dispatch_response
 
 					for(my $i = 0; $i < $count; $i++)
 					{
-						push(@{$return_arr[$iter]}, (($dummy_bool_array_bits[floor($i / 8)] & (1 << ($i % 8))) != 0));
+						push(@{$return_arr[$iter]}, 0 + (($dummy_bool_array_bits[floor($i / 8)] & (1 << ($i % 8))) != 0));
 					}
 
 					$iter++;
@@ -1639,7 +1658,7 @@ sub _dispatch_response
 
 			if(defined($fid))
 			{
-				$self->_dispatch_callback($uid, $fid, \@return_arr);
+				$self->_dispatch_callback($uid, $fid, \@return_arr, 0);
 			}
 
 			return @return_arr;
@@ -1669,7 +1688,7 @@ sub _dispatch_response
 
 			if(defined($fid))
 			{
-				$self->_dispatch_callback($uid, $fid, \@return_arr);
+				$self->_dispatch_callback($uid, $fid, \@return_arr, 0);
 			}
 
 			return @return_arr;
@@ -1685,19 +1704,20 @@ sub _dispatch_response
 			my @unpack_tmp_arr = unpack($form_unpack_patched, $payload);
 
 			if($form_unpack_arr_0_arr[0] eq '?') {
-				my ($count) = $form_unpack_arr[0] =~ s/[^\d]//g;
+				$form_unpack_arr[0] =~ s/[^\d]//g;
+				my ($count) = $form_unpack_arr[0];
 				my @return_arr = (0) x $count;
 				my $n = ceil($count / 8);
 
 				my @payload_ = unpack('(C'.$n.')<', $payload);
 
 				for(my $i = 0; $i < $count; $i++) {
-					$return_arr[$i] = (($payload_[floor($i / 8)] & (1 << ($i % 8))) != 0);
+					$return_arr[$i] = 0 + (($payload_[floor($i / 8)] & (1 << ($i % 8))) != 0);
 				}
 
 				if(defined($fid))
 				{
-					$self->_dispatch_callback($uid, $fid, \@return_arr);
+					$self->_dispatch_callback($uid, $fid, \@return_arr, 1);
 				}
 
 				return \@return_arr;
@@ -1708,7 +1728,7 @@ sub _dispatch_response
 
 				if(defined($fid))
 				{
-					$self->_dispatch_callback($uid, $fid, \@return_arr);
+					$self->_dispatch_callback($uid, $fid, \@return_arr, 1);
 				}
 
 				return \@return_arr;
@@ -1719,7 +1739,7 @@ sub _dispatch_response
 
 				if(defined($fid))
 				{
-					$self->_dispatch_callback($uid, $fid, $return_arr[0]);
+					$self->_dispatch_callback($uid, $fid, $return_arr[0], 1);
 				}
 
 				return $return_arr[0];
@@ -1730,7 +1750,7 @@ sub _dispatch_response
 
 				if(defined($fid))
 				{
-					$self->_dispatch_callback($uid, $fid, \@return_arr);
+					$self->_dispatch_callback($uid, $fid, \@return_arr, 1);
 				}
 
 				return \@return_arr;
@@ -1742,7 +1762,7 @@ sub _dispatch_response
 
 			if(defined($fid))
 			{
-				$self->_dispatch_callback($uid, $fid, $unpack_tmp);
+				$self->_dispatch_callback($uid, $fid, $unpack_tmp, 1);
 			}
 
 			return $unpack_tmp;
@@ -1752,7 +1772,7 @@ sub _dispatch_response
 	{
 		if(defined($fid))
 		{
-			$self->_dispatch_callback($uid, $fid, undef);
+			$self->_dispatch_callback($uid, $fid, undef, 1);
 		}
 
 		return 1;
@@ -1763,7 +1783,7 @@ sub _dispatch_response
 
 sub _dispatch_callback
 {
-	my ($self, $uid, $fid, $args, $is_array) = @_;
+	my ($self, $uid, $fid, $args, $args_is_single_value) = @_;
 
 	if(defined($uid))
 	{
@@ -1873,34 +1893,46 @@ sub _dispatch_callback
 
 				if(defined($self->{devices}->{$uid}->{registered_callbacks}->{$fid}))
 				{
-					# FIXME: for some unknown reason directly passing the reference
-					#        to an array doesn't work with eval. it only works if the
-					#        reference is created locally by eval
-					my @array = @{$args};
-					eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\@array)");
+					if($args_is_single_value)
+					{
+						# pass args as array reference
+						eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\$args)");
+					}
+					else
+					{
+						# pass args as array value to allow for proper struct-like unpacking in the callback
+						my @array = @{$args};
+						eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}(\@array)");
+					}
 				}
 			}
-			else
+			elsif(defined($self->{devices}->{$uid}->{registered_callbacks}->{$fid}))
 			{
 				eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}($args)");
 			}
 		}
-		else
+		elsif(defined($self->{devices}->{$uid}->{registered_callbacks}->{$fid}))
 		{
 			eval("$self->{devices}->{$uid}->{registered_callbacks}->{$fid}()");
 		}
 	}
-	else
+	elsif(defined($self->{registered_callbacks}->{$fid}))
 	{
 		if(defined($args))
 		{
 			if(ref($args) eq "ARRAY")
 			{
-				# FIXME: for some unknown reason directly passing the reference
-				#        to an array doesn't work with eval. it only works if the
-				#        reference is created locally by eval
-				my @array = @{$args};
-				eval("$self->{registered_callbacks}->{$fid}(\@array)");
+				if ($args_is_single_value)
+				{
+					# pass args as array reference
+					eval("$self->{registered_callbacks}->{$fid}(\$args)");
+				}
+				else
+				{
+					# pass args as array value to allow for proper struct-like unpacking in the callback
+					my @array = @{$args};
+					eval("$self->{registered_callbacks}->{$fid}(\@array)");
+				}
 			}
 			else
 			{
@@ -1929,6 +1961,11 @@ sub _dispatch_packet
 	{
 		if(defined($self->{registered_callbacks}->{&CALLBACK_ENUMERATE}))
 		{
+			if (length($packet) != 34)
+			{
+				return 1; # silently ignoring callback with wrong length
+			}
+
 			$self->_dispatch_response($payload, 'Z8 Z8 Z C3 C3 S C', undef, &CALLBACK_ENUMERATE);
 		}
 
@@ -1943,7 +1980,23 @@ sub _dispatch_packet
 	if(defined($self->{devices}->{$uid}->{registered_callbacks}->{$fid}) ||
 	   defined($self->{devices}->{$uid}->{high_level_callbacks}->{-$fid}))
 	{
-		$self->_dispatch_response($payload, $self->{devices}->{$uid}->{callback_formats}->{$fid}, $uid, $fid);
+		eval
+		{
+			$self->{devices}->{$uid}->_check_validity();
+		}
+		or do
+		{
+			return 1; # silently ignoring callbacks from mismatching devices
+		};
+
+		my $format = $self->{devices}->{$uid}->{callback_formats}->{$fid};
+
+		if (length($packet) != @{$format}[0])
+		{
+			return 1; # silently ignoring callback with wrong length
+		}
+
+		$self->_dispatch_response($payload, @{$format}[1], $uid, $fid);
 	}
 
 	return 1;
